// Class CDatabase64, Version 1.2 Feb 2013
//
// Version History
//
// Version 1.2, Feb 2013.	Small (esthetic) changes
// Version 1.1, Jun 2011.	Changes due to new DFLD definition.
// Version 1.0,				64-Bit Implementation
//

#include <afxwin.h>
#include <afx.h>

#include <math.h>		// pow 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "CEngineDef.h"
#include "CEngineDefinitions.h"

#include "CMoveGen64.h"
#include "CDatabase64.h"


DBINDEX MULT[64][64] ;
DBINDEX iIndex[6][64] ;

unsigned int upow3[5] = { 1, 3, 9, 27, 81 } ;


CDatabase64::CDatabase64()
{

	Database64_Init() ;

	InitializeCriticalSection(&m_DB64Lock) ;

}


CDatabase64::~CDatabase64()
{

}


void CDatabase64::Database64_Init(void)
{
	int i ;
	DWORD uIndex ;
	BITBOARD bbBitBoard ;

	for ( i=0; i<64; i++ ) {

		bbBitBoard = (BITBOARD)1<<i ;

		if ( bbBitBoard & DFLD_ROW1 ) {				// Row 1
			m_iWhiteSlice[i] = 40 ;
			m_iBlackSlice[i] = 5 ;

		} else if ( bbBitBoard & DFLD_ROW2 ) {		// Row 2
			m_iWhiteSlice[i] = 35 ;
			m_iBlackSlice[i] = 10 ;

		} else if ( bbBitBoard & DFLD_ROW3 ) {		// Row 3
			m_iWhiteSlice[i] = 30 ;
			m_iBlackSlice[i] = 15 ;
	
		} else if ( bbBitBoard & DFLD_ROW4 ) {		// Row 4
			m_iWhiteSlice[i] = 25 ;
			m_iBlackSlice[i] = 20 ;
		
		} else if ( bbBitBoard & DFLD_ROW5 ) {		// Row 5
			m_iWhiteSlice[i] = 20 ;
			m_iBlackSlice[i] = 25 ;
			
		} else if ( bbBitBoard & DFLD_ROW6 ) {		// Row 6
			m_iWhiteSlice[i] = 15 ;
			m_iBlackSlice[i] = 30 ;
				
		} else if ( bbBitBoard & DFLD_ROW7 ) {		// Row 7
			m_iWhiteSlice[i] = 10 ;
			m_iBlackSlice[i] = 35 ;
					
		} else if ( bbBitBoard & DFLD_ROW8 ) {		// Row 8
			m_iWhiteSlice[i] = 5 ;
			m_iBlackSlice[i] = 40 ;

		} else {
			m_iWhiteSlice[i] = 0 ;
			m_iBlackSlice[i] = 0 ;
		}
	}

	for ( i=0; i<64; i++ )	// Clear m_cix
		m_cix[i] = 0 ;

	for ( i=0; i<5; i++ )	// Clear m_cxi
		m_cxi[i] = 0 ;

	// Method for calculating m_cix and m_cxi

	for ( i=1; i<=50; i++ ) {

		BITSCANFORWARD64(&uIndex, bbDFLD[i]) ;			// Position i
		m_cix[uIndex]	= i ;
		m_cxi[i]		= uIndex ;
	}
	
}


DBINDEX CDatabase64::BinomialCoefficient(int n, int k)
{
	DBINDEX i, iBinomCoef = 1 ;


	if ( k < 0 ) return ( 0  );

	for ( i = k+1;	i <= n;		i++ ) iBinomCoef *= i ;
	for ( i = 2;	i <= (n-k); i++ ) iBinomCoef /= i ;

	return ( iBinomCoef ) ;
}


void CDatabase64::Init_Index64()
{
	int i, j, i0 ;

    
	for ( i=0; i<=50; i++ ) {
    
		i0 = i-10 ;
        
		if ( i0 < 0 ) i0=0 ;
        
		for ( j=0;  j<12; j++ ) MULT[i][j] = 0 ;
        for ( j=i0; j<=i; j++ ) MULT[i][j+12] = BinomialCoefficient(i, j) ;
    }

    for ( i=0; i<=50; i++ ) {

		iIndex[0][i] = 1 ;
		iIndex[1][i] = i ;
		iIndex[2][i] = i * (i-1) / 2 ;
		iIndex[3][i] = i * (i-1) * (i-2) / 6 ;
		iIndex[4][i] = i * (i-1) * (i-2) * (i-3) / 24 ;
		iIndex[5][i] = i * (i-1) * (i-2) * (i-3) * (i-4) / 120 ;
	}
}


DBINDEX CDatabase64::Index64(PTSTACK pTreeStack)
{
	int nr_of_white_man = 0 ;
	int nr_of_black_man = 0 ;
	int nr_of_white_king = 0 ;
	int nr_of_black_king = 0 ;
	int iPosition ;
	unsigned long uIndex ;
	DBINDEX p1, p2, p3 ;
	DBINDEX iIndexWhiteMan = 0, iIndexWhiteKing = 0 , iIndexBlackMan = 0, iIndexBlackKing = 0 ;
	DBINDEX iDBIndex ;
	BITBOARD bbWhiteMan, bbWhiteKing, bbBlackMan, bbBlackKing ;


	bbWhiteMan	= pTreeStack->bbField[BB_WHITEMAN] ;
	bbWhiteKing = pTreeStack->bbField[BB_WHITEKING] ;
	bbBlackMan	= pTreeStack->bbField[BB_BLACKMAN] ;
	bbBlackKing = pTreeStack->bbField[BB_BLACKKING] ;
	
	if ( pTreeStack->bTurn == true ) {			// White to Move
		
		while ( bbWhiteMan ) {

			BITSCANREVERSE64(&uIndex, bbWhiteMan) ;

			bbWhiteMan	^= (BITBOARD)1<<uIndex ;
			iPosition	 = m_cix[uIndex] ;

			iIndexWhiteMan += iIndex[++nr_of_white_man][50 - iPosition] ;
		}

		while ( bbWhiteKing ) {

			BITSCANREVERSE64(&uIndex, bbWhiteKing) ;

			bbWhiteKing	^= (BITBOARD)1<<uIndex ;
			iPosition	 = m_cix[uIndex] ;

			iIndexWhiteKing += iIndex[++nr_of_white_king][50 - iPosition] ;
		}

		while ( bbBlackMan ) {

			BITSCANFORWARD64(&uIndex, bbBlackMan) ;

			bbBlackMan	^= (BITBOARD)1<<uIndex ;
			iPosition	 = m_cix[uIndex] - 1 ;

			iIndexBlackMan += iIndex[++nr_of_black_man][iPosition] ;
		}

		while ( bbBlackKing ) {

			BITSCANFORWARD64(&uIndex, bbBlackKing) ;

			bbBlackKing	^= (BITBOARD)1<<uIndex ;
			iPosition	 = m_cix[uIndex] - 1 ;

			iIndexBlackKing += iIndex[++ nr_of_black_king][iPosition] ;
		}
	
	} else {									// Black to Move
			
		while ( bbWhiteMan ) {

			BITSCANREVERSE64(&uIndex, bbWhiteMan) ;

			bbWhiteMan	^= (BITBOARD)1<<uIndex ;
			iPosition	 = m_cix[uIndex] ;

			iIndexBlackMan += iIndex[++nr_of_black_man][50 - iPosition] ;
		}

		while ( bbWhiteKing ) {

			BITSCANREVERSE64(&uIndex, bbWhiteKing) ;

			bbWhiteKing	^= (BITBOARD)1<<uIndex ;
			iPosition	 = m_cix[uIndex] ;

			iIndexBlackKing += iIndex[++nr_of_black_king][50 - iPosition] ;
		}

		while ( bbBlackMan ) {

			BITSCANFORWARD64(&uIndex, bbBlackMan) ;

			bbBlackMan	^= (BITBOARD)1<<uIndex ;
			iPosition	 = m_cix[uIndex] - 1 ;

			iIndexWhiteMan += iIndex[++nr_of_white_man][iPosition] ;
		}

		while ( bbBlackKing ) {

			BITSCANFORWARD64(&uIndex, bbBlackKing) ;

			bbBlackKing	^= (BITBOARD)1<<uIndex ;
			iPosition	 = m_cix[uIndex] - 1 ;

			iIndexWhiteKing += iIndex[++nr_of_white_king][iPosition] ;
		}
	}
    
	p3 = MULT[50][50- nr_of_white_king + 12] ;
	p2 = MULT[50][50- nr_of_black_king + 12] * p3 ;
	p1 = MULT[45][45 - nr_of_white_man + 12] * p2 ;

	iDBIndex = ( iIndexBlackMan * p1 ) + ( iIndexWhiteMan * p2 ) + ( iIndexBlackKing * p3 ) + iIndexWhiteKing ;

	return ( iDBIndex ) ;
}


char* CDatabase64::cDatabaseName(int iXWhiteMan, int iXWhiteKing, int iXBlackMan, int iXBlackKing, int iWhiteSlice, int iBlackSlice)

{
	static char cDBName[40] ;
	int a=0, i ;

	cDBName[a++] = 'w' ;
	cDBName[a++] ='d';
	cDBName[a++] ='l';
	cDBName[a++] = 48 + iXWhiteMan + iXWhiteKing + iXBlackMan + iXBlackKing ;
	cDBName[a++] = '-' ;
	
	for( i=0; i<iXWhiteKing; i++ )		cDBName[a++] = 'X' ;
	for( i=0; i<iXWhiteMan; i++ )		cDBName[a++] = 'O' ;

	cDBName[a++] = 'v' ;

	for( i=0; i<iXBlackKing; i++ )		cDBName[a++] = 'X' ;
	for( i=0; i<iXBlackMan; i++ )		cDBName[a++] = 'O' ;

	cDBName[a++]	= '-' ;
	cDBName[a++]	= 48 + iWhiteSlice ;
	cDBName[a++]	= 48 + iBlackSlice ;
	cDBName[a]		= '\0';
 
	return ( cDBName ) ;
}


void CDatabase64::DB64_Init(PENGINEDATA pEngineData)
{
	bool bDB64Cache_Init ;
	CString csMessage ;


	Init_Index64() ;

	// DBCache Routines
	bDB64Cache_Init = DB64Cache_Init(pEngineData) ;

	if ( bDB64Cache_Init == true )
		csMessage.Format("DB64 Cache Init OK , %d 4KB Blocks Allocated", DB64Cache.iDBCacheSize) ;	
	else
		csMessage.Format("DB64 Cache Init Error") ;	

	SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage ) ;

	DB64Cache_Open(DB64_COMPRESS_DEFAULT) ;

	csMessage.Format("DB64 File Open = %d", DB64Cache.iNFilesOpen ) ;	
	SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage ) ;

	csMessage.Format("DB64 ReadData Index : Start" ) ;
	SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage ) ;

	DB64Cache_ReadDataIndex(DB64_COMPRESS_DEFAULT) ;

	csMessage.Format("DB64 ReadData Index : End" ) ;	
	SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage ) ;

}


// Returns the ID of the requested database
int CDatabase64::DatabaseID(int color, int iXWhiteMan, int iXWhiteKing, int iXBlackMan, int iXBlackKing, int iWhiteSlice, int iBlackSlice)
{
	int iDBID ;

	if ( color == DB64_WHITE ) 
		iDBID = 81 * ( iXWhiteMan + 8*iXWhiteKing + 64*iXBlackMan + 512*iXBlackKing ) + 9*iWhiteSlice + iBlackSlice ;
	else 
		iDBID = 81 * ( iXBlackMan + 8*iXBlackKing + 64*iXWhiteMan + 512*iXWhiteKing ) + 9*iBlackSlice + iWhiteSlice ;
	
	return ( iDBID ) ;
}


// Returns the ID of the requested database
int CDatabase64::DatabaseID(PTSTACK pTDepth)
{
	int iDBID ;

	if ( pTDepth->bTurn == true ) 
		iDBID = 81 * ( pTDepth->iWMan + 8*pTDepth->iWKing + 64*pTDepth->iBMan + 512*pTDepth->iBKing ) ;
	else 
		iDBID = 81 * ( pTDepth->iBMan + 8*pTDepth->iBKing +64*pTDepth->iWMan + 512*pTDepth->iWKing ) ;
	
	return ( iDBID ) ;
}  


void CDatabase64::SendEngineMessage(int iMessageSource, PENGINEDATA pEngineData, int iMessageID, CString csMessage)
{
	int iIndex ;

	pEngineData->SendEngineMessageLock.Lock() ;

	iIndex = pEngineData->iOutputStackWrite % ENGINE_OUTPUT_STACK_SIZE ;

	pEngineData->OutputStack[iIndex].iMessageSource		= iMessageSource ;
	pEngineData->OutputStack[iIndex].iMessageID			= iMessageID ;
	pEngineData->OutputStack[iIndex].csOutputMessage	= csMessage ;
	
	++(pEngineData->iOutputStackWrite) ;

	pEngineData->SendEngineMessageLock.Unlock() ;
}


// Cache Functions
bool CDatabase64::DB64Cache_FENRead(char* cBuf, PDB64POSITION pDB64Position)
{
	char cChar ;
	int iColor, iDigit0, iDigit1, iFen, iLength, iPiece, iPosition ;
	int iNWhiteMan, iNWhiteKing, iNBlackMan, iNBlackKing ;

	if ( cBuf[0] == 'W' )
		pDB64Position->bTurn = true ;
	else if ( cBuf[0] == 'B' )
		pDB64Position->bTurn = false ;
	else
		return ( false) ;		// FEN should start with W or B

	if ( cBuf[1] != ':' )
		return (false ) ;		// FEN[1] should contain ':' character

	iLength = (int)strlen(cBuf) ;

	if ( cBuf[iLength-2] != '.' )
		return ( false ) ;		// FEN should end with '.' character

	iFen	= 2 ;
	iColor	= -1 ;				// Non-Existing Color

	iNWhiteMan	= 0 ;
	iNWhiteKing = 0 ;
	iNBlackMan	= 0 ;
	iNBlackKing = 0 ;

	pDB64Position->bbField[BB_WHITEMAN]		= 0 ;	// Init BitFields
	pDB64Position->bbField[BB_WHITEKING]	= 0 ;
	pDB64Position->bbField[BB_BLACKMAN]		= 0 ;
	pDB64Position->bbField[BB_BLACKKING]	= 0 ;

	do {

		cChar = cBuf[iFen] ;

		if ( cChar == 'W' ) {

			iColor	= WHITE ;
			iPiece = BB_WHITEMAN ;

		} else if ( cChar == 'B' ) {

			iColor = BLACK ;
			iPiece = BB_BLACKMAN ;

		} else if ( cChar == 'K' ) {

			if ( iColor == WHITE )
				iPiece = BB_WHITEKING ;
			else if ( iColor == BLACK ) 
				iPiece = BB_BLACKKING ;
			else
				return ( false ) ;	// Piece identification can not start without Color Defined 
		
		} else if ( cChar == ',' ) {

			if ( iColor == WHITE )
				iPiece = BB_WHITEMAN ;
			else if ( iColor == BLACK )
				iPiece = BB_BLACKMAN ;
			else
				return (false) ;	// Character ',' can not occur before Color Defined 
		
		} else if ( isdigit(cChar) ) {

			iDigit0 = atoi(&cChar) ;

			cChar = cBuf[iFen+1] ;

			if ( isdigit(cChar) ) {		// Position has 2 characters

				iDigit1		= atoi(&cChar) ;
				iPosition	= ( 10 * iDigit0 ) + iDigit1 ;
				
				++iFen ;

			} else {					// Position has 1 character

				iPosition = iDigit0 ;
			}

			pDB64Position->bbField[iPiece] |= (BITBOARD)1<<m_cxi[iPosition] ;

			if ( iPiece == BB_WHITEMAN )
				++iNWhiteMan ;
			else if ( iPiece == BB_WHITEKING )
				++iNWhiteKing ;
			else if ( iPiece == BB_BLACKMAN )
				++iNBlackMan ;
			else if ( iPiece == BB_BLACKKING )
				++iNBlackKing ;
		}

		++iFen ;

	} while ( cBuf[iFen] != '.' ) ;

	pDB64Position->iNWhiteMan	= iNWhiteMan ;
	pDB64Position->iNWhiteKing	= iNWhiteKing ;
	pDB64Position->iNBlackMan	= iNBlackMan ;
	pDB64Position->iNBlackKing	= iNBlackKing ;

	return ( true ) ;
}


int CDatabase64::DB64Cache_TestRead(PDB64POSITION pDB64Position)
{
	char cBuf[64] ;
	int iCount, iNError ;
	CString csFileName ;
	FILE *fp ;

	iCount	= 0 ;
	iNError	= 0 ;

	csFileName = m_csDatabasePath + DB64_TESTFILE ;

	if ( fopen_s(&fp, csFileName, "r") != 0 )
		return ( iCount ) ;

	while ( fgets(cBuf, sizeof(cBuf), fp) ) {

		if ( DB64Cache_FENRead(cBuf, &pDB64Position[iCount]) == false )
			++iNError ;

		++iCount ;
	}

	fclose(fp) ;

	return ( iCount ) ;
}


bool CDatabase64::DB64Cache_Test(PENGINEDATA pEngineData, int iDB64CompressMode)
{
	int i, j, k, iCount, iDatabaseScore ;
	float fStartTime, fEndTime ;
	CString csMessage ;
	DBINDEX iNWin, iNDraw, iNLose ;
	PTSTACK pTDepth ;
	PDB64POSITION pDB64Position ;


	pTDepth			= pEngineData->SearchThreadData[0].pTreeStack ;
	pDB64Position	= (PDB64POSITION)calloc(DB64_MAXTESTPOSITIONS, sizeof(DB64POSITION) ) ;

	if ( pDB64Position == NULL )	// No Memory
		return ( false ) ;

	iCount = DB64Cache_TestRead(pDB64Position) ;

	iCount = min(iCount, DB64_MINTESTPOSITIONS) ;

	if ( iCount == 0 )			// Error File Read
		return ( false ) ;

	for ( k=0; k<5; k++ ) {		// Do test 5 x for statistics

		for ( i=0; i<DB64Cache.iDBIndexTable; i++ )			// Clear Cache
			DB64Cache.pDBCacheIDTable[i] = DB64CACHE_BLOCK_NOTLOADED ;
	
		DB64Cache.iNCacheRead		= 0 ;
		DB64Cache.iNFilesOpen		= 0 ;
		DB64Cache.iNCacheWrite		= 0 ;
		DB64Cache.iCacheFreeBlocks	= DB64Cache.iDBCacheSize ;
		DB64Cache.iCacheLRU			= 0 ;
		DB64Cache.iCacheMRU			= 0 ;

		iNWin	= 0 ;
		iNDraw	= 0 ;
		iNLose	= 0 ;

		fStartTime = (float)clock() ;	// Test with empty DataBase

		for ( j=0; j<iCount; j++ ) {

			pTDepth->bTurn = pDB64Position[j].bTurn	;	// Copy Position

			pTDepth->bbField[BB_WHITEMAN]	= pDB64Position[j].bbField[BB_WHITEMAN] ;
			pTDepth->bbField[BB_WHITEKING]	= pDB64Position[j].bbField[BB_WHITEKING] ;
			pTDepth->bbField[BB_BLACKMAN]	= pDB64Position[j].bbField[BB_BLACKMAN] ;
			pTDepth->bbField[BB_BLACKKING]	= pDB64Position[j].bbField[BB_BLACKKING] ;

			pTDepth->iWMan	= pDB64Position[j].iNWhiteMan ;
			pTDepth->iWKing	= pDB64Position[j].iNWhiteKing	;
			pTDepth->iBMan	= pDB64Position[j].iNBlackMan ;
			pTDepth->iBKing	= pDB64Position[j].iNBlackKing	;

			iDatabaseScore = DB64Cache_Read(pTDepth, iDB64CompressMode) ;

			if ( iDatabaseScore == DB64_WIN )
				++iNWin ;
			else if ( iDatabaseScore == DB64_DRAW )
				++iNDraw ;
			else if ( iDatabaseScore == DB64_LOSE )
				++iNLose ;
		}

		fEndTime = (float)clock() ;

		csMessage.Format("%s Time: %.1f, WDL = %I64d %I64d %I64d", ENGINE_STR_ASX_DB64, (fEndTime-fStartTime)/CLOCKS_PER_SEC , iNWin, iNDraw, iNLose ) ;	
		SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage) ;
			
		csMessage.Format("%s Cache RWF : %I64d %I64d %I64d", ENGINE_STR_ASX_DB64, DB64Cache.iNCacheRead, DB64Cache.iNCacheWrite, DB64Cache.iCacheFreeBlocks ) ;
		SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage) ;
	}

	for ( k=0; k<5; k++ ) {			// Do test 5 x for statistics

		fStartTime = (float)clock() ;	// Test with filled DataBase

		for ( i=0; i<100; i++ ) {

			DB64Cache.iNCacheRead	= 0 ;
			DB64Cache.iNCacheWrite	= 0 ;

			iNWin	= 0 ;
			iNDraw	= 0 ;
			iNLose	= 0 ;

			for ( j=0; j<iCount; j++ ) {

				pTDepth->bTurn = pDB64Position[j].bTurn	;	// Copy Position

				pTDepth->bbField[BB_WHITEMAN]	= pDB64Position[j].bbField[BB_WHITEMAN] ;
				pTDepth->bbField[BB_WHITEKING]	= pDB64Position[j].bbField[BB_WHITEKING] ;
				pTDepth->bbField[BB_BLACKMAN]	= pDB64Position[j].bbField[BB_BLACKMAN] ;
				pTDepth->bbField[BB_BLACKKING]	= pDB64Position[j].bbField[BB_BLACKKING] ;

				pTDepth->iWMan	= pDB64Position[j].iNWhiteMan ;
				pTDepth->iWKing	= pDB64Position[j].iNWhiteKing	;
				pTDepth->iBMan	= pDB64Position[j].iNBlackMan ;
				pTDepth->iBKing	= pDB64Position[j].iNBlackKing	;

				iDatabaseScore = DB64Cache_Read(pTDepth, DB64_COMPRESS_DEFAULT) ;
				
				if ( iDatabaseScore == DB64_WIN )
					++iNWin ;
				else if ( iDatabaseScore == DB64_DRAW )
					++iNDraw ;
				else if ( iDatabaseScore == DB64_LOSE )
					++iNLose ;
			}
		}

		fEndTime = (float)clock() ;

		csMessage.Format("%s Time: %.1f, WDL = %I64d %I64d %I64d", ENGINE_STR_ASX_DB64, (fEndTime-fStartTime)/CLOCKS_PER_SEC, iNWin, iNDraw, iNLose ) ;	
		SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage) ;

		csMessage.Format("%s Cache RWF = %I64d %I64d %I64d", ENGINE_STR_ASX_DB64, DB64Cache.iNCacheRead, DB64Cache.iNCacheWrite, DB64Cache.iCacheFreeBlocks ) ;
		SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage) ;
	}

	return ( true ) ;
}


bool CDatabase64::DB64Cache_Open(int iDB64CompressMode)

{
	bool bDB64InfoWrite, bDB64InfoRead ;
	char *cDBName ;
	int i, iNPiece, iNWhitePiece, iNWhiteMan, iNWhiteKing, iNBlackMan, iNBlackKing ;
	int iDBID ;
	__int64 iNCacheBlocks ;
	CString csInfoFileName, csDBFileName ;

	DB64HEADER DB64Header ;
	DWORD dwBytesRead ;
	HANDLE hFile, hFileInfo ;

	bDB64InfoWrite			= false ;
	bDB64InfoRead			= true ;
	iNCacheBlocks			= 0 ;

	DB64Cache.iDBIndexTable = iNCacheBlocks ;
	DB64Cache.iNFilesOpen	= 0 ;

	// Init arrays
	for ( i=0; i<(4096 * 81); i++ ) {

		DB64Cache.iDB64BlockStart[i]	= DB64CACHE_ILLEGALBLOCK ;
		DB64Cache.hFileDB64[i]			= INVALID_HANDLE_VALUE ;
	}

	csInfoFileName = m_csDatabasePath + DB64_INFOFILE ;

	if ( bDB64InfoWrite | bDB64InfoRead )
		hFileInfo = CreateFile(csInfoFileName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL ) ;

	for ( iNPiece=2; iNPiece<=DB64CACHE_MAXPIECE; iNPiece++ )
		for ( iNWhitePiece=1; iNWhitePiece<=(iNPiece-1); iNWhitePiece++ )
			for ( iNWhiteKing=0; iNWhiteKing<=iNWhitePiece; iNWhiteKing++ )
				for ( iNBlackKing=0; iNBlackKing<=(iNPiece-iNWhitePiece); iNBlackKing++ )	{

					iNWhiteMan = iNWhitePiece - iNWhiteKing ;
					iNBlackMan = iNPiece - iNWhitePiece - iNBlackKing ;

					iDBID = DatabaseID(DB64_WHITE, iNWhiteMan, iNWhiteKing, iNBlackMan, iNBlackKing, 0, 0) ;

					cDBName = cDatabaseName(iNWhiteMan, iNWhiteKing, iNBlackMan, iNBlackKing, 0, 0) ;
	
					csDBFileName.Format("%s%s.cprz", m_csDatabasePath, cDBName) ;

					hFile = CreateFile(csDBFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_READONLY | FILE_FLAG_NO_BUFFERING, NULL ) ;
//					hFile = CreateFile(cFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_READONLY, NULL ) ;

					if ( hFile != INVALID_HANDLE_VALUE ) {

						DB64Cache.hFileDB64[iDBID] = hFile ;
						++DB64Cache.iNFilesOpen ;

						if ( bDB64InfoRead == false ) {

							ReadFile(hFile, &DB64Header, sizeof(DB64HEADER), &dwBytesRead, NULL ) ;

							ASSERT ( iDBID == DB64Header.iDBID ) ;

							if ( bDB64InfoWrite )
								WriteFile(hFileInfo, &DB64Header, sizeof(DB64HEADER), &dwBytesRead, NULL ) ;

	
						} else {

							ReadFile(hFileInfo, &DB64Header, sizeof(DB64HEADER), &dwBytesRead, NULL ) ;

							iDBID = (int)DB64Header.iDBID ;
						}
	
						DB64Cache.DB64Header[iDBID] = DB64Header ;

						iNCacheBlocks += DB64Header.iNBlocks ;
					}
				}

	DB64Cache.iDBIndexTable = iNCacheBlocks ;

	if ( bDB64InfoWrite | bDB64InfoRead )
		CloseHandle( hFileInfo ) ;

	return ( true ) ;
}


bool CDatabase64::DB64Cache_Close(void)

{
	int i ;


	for ( i=0; i<(4096 * 81); i++ ) {

		if ( DB64Cache.hFileDB64[i] != INVALID_HANDLE_VALUE )
			CloseHandle( DB64Cache.hFileDB64[i] ) ;
	}

	return ( true ) ;
}

bool CDatabase64::DB64Cache_Clear(PENGINEDATA pEngineData)

{
	int i ;
	int *pDBCacheIDTable ;

	DBINDEX *pDBCachePrevBlock, *pDBCacheNextBlock, *pDBCacheIndexTable ;


	pDBCacheIDTable		= DB64Cache.pDBCacheIDTable ;
	pDBCacheIndexTable	= DB64Cache.pDBCacheIndexTable ;
	pDBCachePrevBlock	= DB64Cache.pDBCachePrevBlock ;
	pDBCacheNextBlock	= DB64Cache.pDBCacheNextBlock ;

	for ( i=0; i<DB64Cache.iDBIndexTable; i++ )
		pDBCacheIDTable[i] = DB64CACHE_BLOCK_NOTLOADED ;

	for ( i=0; i<DB64Cache.iDBCacheSize; i++ ) {

		pDBCacheIndexTable[i]	= DB64CACHE_ILLEGALINDEX ;

		pDBCachePrevBlock[i]	= DB64CACHE_ILLEGALBLOCK ;
		pDBCacheNextBlock[i]	= DB64CACHE_ILLEGALBLOCK ;
	}

	DB64Cache.iNCacheRead			= 0 ;
	DB64Cache.iNCacheHeuristic		= 0 ;
	DB64Cache.iNCacheHeuristicError	= 0 ;
	DB64Cache.iNCacheWrite			= 0 ;
	DB64Cache.iCacheFreeBlocks		= DB64Cache.iDBCacheSize ;
	DB64Cache.iCacheLRU				= 0 ;
	DB64Cache.iCacheMRU				= 0 ;

	return ( true ) ;
}


bool CDatabase64::DB64Cache_Init(PENGINEDATA pEngineData)
{
	unsigned char *pDBCache, *pDBCache0 ;
	int i, iDBCacheSize ;
	DBINDEX *pDBCachePrevBlock, *pDBCacheNextBlock, *pDBCacheIndexTable ;


	iDBCacheSize = (int)pow((double)2,( pEngineData->iDBCacheSize ) ) ;

	DB64Cache.iDBCacheSize = iDBCacheSize ;

	// Allocate Memory for Cache, CacheIndexTable and Double-Linked List
	pDBCache0 = (unsigned char*)calloc( iDBCacheSize + 1, DB64CACHE_BLOCKSIZE ) ;		// Add 1 ( 4KByte ) Block for allignment
	
	if ( pDBCache0 == NULL )
		return ( false ) ;

	if ( (unsigned __int64)pDBCache0 % DB64CACHE_BLOCKSIZE  == 0 )						// Allign pointer to 4KByte Block boundary
		pDBCache = pDBCache0 ;
	else
		pDBCache = pDBCache0 + ( DB64CACHE_BLOCKSIZE - ( (unsigned __int64)pDBCache0 % DB64CACHE_BLOCKSIZE ) ) ;

	pDBCacheIndexTable = (DBINDEX*)calloc( iDBCacheSize, sizeof(DBINDEX) ) ;

	if ( pDBCacheIndexTable == NULL )
		return ( false ) ;

	pDBCachePrevBlock = (DBINDEX*)calloc( iDBCacheSize, sizeof(DBINDEX) ) ;

	if ( pDBCachePrevBlock == NULL )
		return ( false ) ;

	pDBCacheNextBlock = (DBINDEX*)calloc( iDBCacheSize, sizeof(DBINDEX) ) ;

	if ( pDBCacheNextBlock == NULL )
		return ( false ) ;

	for ( i=0; i<iDBCacheSize; i++ ) {

		pDBCacheIndexTable[i]	= DB64CACHE_ILLEGALINDEX ;

		pDBCachePrevBlock[i]	= DB64CACHE_ILLEGALBLOCK ;
		pDBCacheNextBlock[i]	= DB64CACHE_ILLEGALBLOCK ;
	}

	DB64Cache.iNCacheRead			= 0 ;
	DB64Cache.iNCacheHeuristic		= 0 ;
	DB64Cache.iNCacheHeuristicError	= 0 ;

	DB64Cache.pDBCache				= pDBCache ;
	DB64Cache.pDBCache0				= pDBCache0 ;
	DB64Cache.pDBCacheIndexTable	= pDBCacheIndexTable ;
	DB64Cache.pDBCachePrevBlock		= pDBCachePrevBlock ;
	DB64Cache.pDBCacheNextBlock		= pDBCacheNextBlock ;

	DB64Cache.iCacheFreeBlocks		= iDBCacheSize ;
	DB64Cache.iCacheLRU				= 0 ;
	DB64Cache.iCacheMRU				= 0 ;

	return ( true ) ;
}


bool CDatabase64::DB64Cache_Modify(PENGINEDATA pEngineData)
{
	bool bDB64Cache_Init ;
	CString csMessage ;


	free( DB64Cache.pDBCacheNextBlock ) ;
	free( DB64Cache.pDBCachePrevBlock ) ;
	free( DB64Cache.pDBCacheIndexTable ) ;
	free( DB64Cache.pDBCache0 ) ;

	if ( pEngineData->iDBCacheSize == DBCACHESIZE_DEFAULT_NOCACHE )
		return ( false ) ;

	bDB64Cache_Init = DB64Cache_Init(pEngineData) ;

	if ( bDB64Cache_Init == true )
		csMessage.Format("DB64 Cache Init OK , %d 4KB Blocks Allocated", DB64Cache.iDBCacheSize) ;	
	else
		csMessage.Format("DB64 Cache Init Error") ;	

	SendEngineMessage(ENGINE_CONSOLE, pEngineData, ENGINE_ASX_DB64, csMessage ) ;

	if ( bDB64Cache_Init == false )
		return (false ) ;

	DB64Cache_Clear(pEngineData) ;

	return ( true ) ;
}


__int64 CDatabase64::DB64Cache_ReadBlock(int iDBID, DBINDEX iBlock)

{
	unsigned char* pucBuffer ;
	DBINDEX iCacheBlockID, iReadOffset ;
	DWORD dwBytesRead ;
	LONG iReadOffsetHigh, iReadOffsetLow ;
	HANDLE hFile ;


	hFile = DB64Cache.hFileDB64[iDBID] ;

	if ( hFile == INVALID_HANDLE_VALUE )
		return ( DB64CACHE_READ_ERROR ) ;

	if ( DB64Cache.iCacheFreeBlocks ) {

		iCacheBlockID	=  ( DB64Cache.iDBCacheSize - DB64Cache.iCacheFreeBlocks ) ;

		--DB64Cache.iCacheFreeBlocks ;

	} else {

		iCacheBlockID	= DB64Cache.iCacheLRU ;

		DB64Cache.pDBCacheIDTable[ DB64Cache.pDBCacheIndexTable[iCacheBlockID] ] = DB64CACHE_BLOCK_NOTLOADED ;
	}

	DB64Cache.pDBCacheIndexTable[iCacheBlockID] = iBlock ;


//	iReadOffset	= sizeof(DB64HEADER) + ( 8 * DB64Cache.DB64Header[iDBID].iNBlocks ) + ( DB64CACHE_BLOCKSIZE * ( iBlock - DB64Cache.iDB64BlockStart[iDBID] ) ) ;		// .cpry
	iReadOffset	= DB64CACHE_BLOCKSIZE * ( iBlock - DB64Cache.iDB64BlockStart[iDBID] ) ;																				// .cprz

	ASSERT ( iReadOffset >= 0 ) ;

	iReadOffsetLow	= iReadOffset & 0xFFFFFFFF ;
	iReadOffsetHigh	= iReadOffset>>32 ;

	SetFilePointer(hFile, iReadOffsetLow, &iReadOffsetHigh, FILE_BEGIN ) ;

	pucBuffer = DB64Cache.pDBCache + ( iCacheBlockID * DB64CACHE_BLOCKSIZE ) ;

	ReadFile(hFile, pucBuffer, DB64CACHE_BLOCKSIZE, &dwBytesRead, NULL ) ;

	DB64Cache.pDBCacheIndexTable[iCacheBlockID] = iBlock ;

	return ( iCacheBlockID ) ;

}


int CDatabase64::DB64Cache_Read(int iDBID, DBINDEX iIndex)

{
	bool bFreeBlock ;
	unsigned char *pcDB64Cache, *pcDB64Cache0 ;
	int iDB64CacheOffset, iDB64Value ;
	DBINDEX iCacheBlockID, iCacheIndex, iCacheBlockIndex, iCacheLRU, iCacheMRU ;
	DBINDEX iBlockID, iBlock, iBlockL, iBlockH, iBlockT ;


	++DB64Cache.iNCacheRead ;

	bFreeBlock = false ;

	iBlock		= ( ( iIndex * DB64Cache.DB64Header[iDBID].iNBlocks ) / DB64Cache.DB64Header[iDBID].iCount ) + DB64Cache.iDB64BlockStart[iDBID] ;
	iBlockID	= iIndex /  DB64CACHE_BLOCKSIZE ;
	iBlockL		= DB64Cache.iDB64BlockStart[iDBID] ;
	iBlockH		= ( DB64Cache.DB64Header[iDBID].iNBlocks - 1 ) + DB64Cache.iDB64BlockStart[iDBID] ;

	while ( DB64Cache.pDBIndexTable[iBlock] < iBlockID || ( iBlock > DB64Cache.iDB64BlockStart[iDBID] ? DB64Cache.pDBIndexTable[iBlock-1] : -1 ) >= iBlockID ) {
									
		if (  DB64Cache.pDBIndexTable[iBlock] < iBlockID ) {
 
			iBlockT	= iBlock ;
			iBlock	= ( iBlock + iBlockH + 1 ) / 2 ;
			iBlockL	= iBlockT + 1 ; 
									
		} else {

			iBlockT	= iBlock ;
			iBlock	= ( iBlock + iBlockL ) / 2 ;
			iBlockH	= iBlockT - 1 ;

		}

		ASSERT( iBlock >= 0 ) ;

	}

	iCacheBlockID = DB64Cache.pDBCacheIDTable[iBlock] ;

	EnterCriticalSection(&m_DB64Lock) ;	// Enter Critical Section ( MP Implementation )

	if ( iCacheBlockID == DB64CACHE_BLOCK_NOTLOADED ) {

		bFreeBlock = ( DB64Cache.iCacheFreeBlocks > 0 ? true : false ) ;

		++DB64Cache.iNCacheWrite ;

		iCacheBlockID = DB64Cache_ReadBlock(iDBID, iBlock) ;

		DB64Cache.pDBCacheIDTable[iBlock] = iCacheBlockID ;
	}
	
	// Update LRU/MRU
	if ( iCacheBlockID != DB64Cache.iCacheMRU ) {	// No update if MRU Block is addressed

		iCacheLRU = DB64Cache.iCacheLRU ;
		iCacheMRU = DB64Cache.iCacheMRU ;

		ASSERT ( iCacheLRU >= 0 ) ;
		ASSERT ( iCacheMRU >= 0 ) ;

		DB64Cache.iCacheMRU							= iCacheBlockID ;
		DB64Cache.pDBCacheNextBlock[iCacheMRU]		= iCacheBlockID ;
		DB64Cache.pDBCachePrevBlock[iCacheBlockID]	= iCacheMRU ;

		if ( iCacheBlockID == iCacheLRU )

			DB64Cache.iCacheLRU = DB64Cache.pDBCacheNextBlock[iCacheLRU] ;
		
		else if ( bFreeBlock == false ) {

			int iDBCachePrevBlock = DB64Cache.pDBCachePrevBlock[iCacheBlockID] ;
			int iDBCacheNextBlock = DB64Cache.pDBCacheNextBlock[iCacheBlockID] ;

			DB64Cache.pDBCacheNextBlock[iDBCachePrevBlock] = iDBCacheNextBlock ;
			DB64Cache.pDBCachePrevBlock[iDBCacheNextBlock] = iDBCachePrevBlock ;
		}
	}

	ASSERT ( iCacheBlockID >= 0 ) ;

	iCacheIndex		= iIndex % DB64CACHE_BLOCKSIZE  ;
	pcDB64Cache0	= &DB64Cache.pDBCache[DB64CACHE_BLOCKSIZE * iCacheBlockID] ;

	if ( iBlock > DB64Cache.iDB64BlockStart[iDBID] )
		iCacheBlockIndex  = 2 * ( iBlockID - ( DB64Cache.pDBIndexTable[iBlock - 1] + 1 ) ) ;
	else
		iCacheBlockIndex = 2 * iBlockID ;

	ASSERT ( iCacheBlockIndex >= 0 ) ;

	iDB64CacheOffset = pcDB64Cache0[iCacheBlockIndex] + 256 * pcDB64Cache0[iCacheBlockIndex + 1] ;

	ASSERT ( iDB64CacheOffset < DB64CACHE_BLOCKSIZE ) ;

	pcDB64Cache	= pcDB64Cache0 + iDB64CacheOffset ;

	iDB64Value = DB64Cache_ReadCPRValue(pcDB64Cache, iCacheIndex) ;

	LeaveCriticalSection(&m_DB64Lock) ;	// Leave Critical Section ( MP Implementation )

	return ( iDB64Value ) ;
}


int CDatabase64::DB64Cache_Read(PTSTACK pTDepth, int iDB64CompressMode)

{
	int iDBID, iDB64Score ;
	DBINDEX iIndex ;

	iDBID = DatabaseID(pTDepth) ;

	if ( DB64Cache.hFileDB64[iDBID] == INVALID_HANDLE_VALUE )
		return ( DB64_UNKNOWN ) ;

	iIndex = Index64(pTDepth) ;

	iDB64Score = DB64Cache_Read(iDBID, iIndex) ;

	return ( iDB64Score ) ;
}


bool CDatabase64::DB64Cache_ReadDataIndex(int iDB64CompressMode)
{
	bool bIndexFile ;
	int iNPiece, iNWhitePiece, iNWhiteMan, iNWhiteKing, iNBlackMan, iNBlackKing ;
	int iDBID, iNFilesRead ;
	int *pDBCacheIDTable ;
	__int64 i, iNCacheBlocks, *pDBIndexTable  ;
	CString csIndexFileName, csPIndexFileName ;
	FILE *fp ;
	DWORD dwBytesRead ;
	HANDLE hFile, hFileIndex ;


	iNCacheBlocks	= 0 ;
	iNFilesRead		= 0 ;

	// Allocate Memory for IndexTable
	pDBIndexTable = (__int64*)calloc( DB64Cache.iDBIndexTable, sizeof(__int64) ) ;

	if ( pDBIndexTable == NULL )
		return ( false ) ;

	// Allocate Memory for Cache IndexTable
	pDBCacheIDTable =  ( int*)calloc( DB64Cache.iDBIndexTable, sizeof(int) ) ;

	if ( pDBCacheIDTable == NULL )
		return ( false ) ;

	bIndexFile = false ;
	for ( i=9; i>=2; i-- ) {

		csIndexFileName.Format("%s%s%d%s", m_csDatabasePath, DB64_INDEXNAME, i, DB64_INDEXEXT) ;

		if ( fopen_s(&fp, csIndexFileName, "r") == 0 ) {

			fclose(fp) ;
			bIndexFile = true ;
			break ;
		}
	}

	if ( bIndexFile == false )
		return ( false ) ;

	DB64Cache.pDBIndexTable		= pDBIndexTable ;
	DB64Cache.pDBCacheIDTable	= pDBCacheIDTable ;

	for ( i=0; i<DB64Cache.iDBIndexTable; i++ )
		pDBCacheIDTable[i] = DB64CACHE_BLOCK_NOTLOADED ;

	hFileIndex = CreateFile(csIndexFileName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL ) ;

	ReadFile(hFileIndex, &pDBIndexTable[0], sizeof(__int64)*DB64Cache.iDBIndexTable, &dwBytesRead, NULL ) ;

	for ( iNPiece=2; iNPiece<=DB64CACHE_MAXPIECE; iNPiece++ ) {
		for ( iNWhitePiece=1; iNWhitePiece<=(iNPiece-1); iNWhitePiece++ ) {
			for ( iNWhiteKing=0; iNWhiteKing<=iNWhitePiece; iNWhiteKing++ ) {
				for ( iNBlackKing=0; iNBlackKing<=(iNPiece-iNWhitePiece); iNBlackKing++ ) {

					iNWhiteMan = iNWhitePiece - iNWhiteKing ;
					iNBlackMan = iNPiece - iNWhitePiece - iNBlackKing ;

					iDBID = DatabaseID(DB64_WHITE, iNWhiteMan, iNWhiteKing, iNBlackMan, iNBlackKing, 0, 0) ;

					hFile = DB64Cache.hFileDB64[iDBID] ;

					if ( hFile != INVALID_HANDLE_VALUE ) {

						DB64Cache.iDB64BlockStart[iDBID] = iNCacheBlocks ;

						iNCacheBlocks += DB64Cache.DB64Header[iDBID].iNBlocks ;
					}	
				}
			}
		}
	}
			
	CloseHandle(hFileIndex) ;

	return ( true ) ;
}


int CDatabase64::DB64Cache_ReadCPRValue(unsigned char* pcDB, DBINDEX iIndex)
{
	unsigned char cByte ;
	DBINDEX i ;

	i = iIndex ;

	do {

		cByte = *pcDB++ ;

		if ( cByte <= 242 ) {
			i -= 5 ;

		} else if ( cByte <= 246 ) {

			if ( cByte == 246 ) i -= *pcDB++ * 5 ;
			else i -= ( cByte - 241 ) * 5 ;

			cByte = 0 ;

		} else if ( cByte <= 250 ) {

			if ( cByte == 250 ) i -= *pcDB++ * 5 ;
			else i -= ( cByte - 245 ) * 5 ;

			cByte = 121 ;

		} else if ( cByte <= 254 ) {

			if ( cByte == 254 ) i -= *pcDB++ * 5 ;
			else i -= ( cByte - 249 ) * 5 ;
				cByte = 242 ;

		} else {

			i -= *pcDB++ * 5 ;
			cByte = *pcDB++ ;
		}

	} while ( i >= 0 ) ;

  return ( ( cByte / upow3[ 4 + ( i + 1 ) % 5 ]) % 3 ) ;

}


bool CDatabase64::DatabasePosition0(PTSTACK pTDepth, int* iScore )
{
	int iValue ;


	iValue = DB64Cache_Read(pTDepth, DB64_COMPRESS_DEFAULT) ;
		
	if ( iValue != DB64_UNKNOWN ) {

		if ( iValue == DB64_WIN )
			*iScore = DATABASE_WIN_VAL  ;

		else if ( iValue == DB64_DRAW )
			*iScore = DRAW_VAL ;

		else if ( iValue == DB64_LOSE )
			*iScore = DATABASE_LOOSE_VAL  ;

		return ( true ) ;

	}	// End Database Lookup

	*iScore = 0 ;

	return ( false ) ;
}


bool CDatabase64::DatabasePosition(bool bDatabasePosition, PTSTACK pTDepth, int* iScore )
{
	bool bCacheHeuristic = false ;
	int iBitScan, iDBID, iDB64Score ;
	int iMatVal, iWhiteSlice, iBlackSlice ;
	DWORD iBitPosition ;


	iDBID = DatabaseID(pTDepth) ;

	if ( DB64Cache.hFileDB64[iDBID] == INVALID_HANDLE_VALUE )
		return ( false ) ;
	
	iDB64Score = DB64Cache_Read(iDBID, Index64(pTDepth)) ;

	if ( iDB64Score == DB64_UNKNOWN )
		return ( false ) ;

	if ( iDB64Score == DB64_DRAW ) {
		pTDepth->uEvalFlags = EVAL_EXACT_DRAW ;
		*iScore = DRAW_VAL ;

		return ( true ) ;
	}

	iBitScan	= BITSCANREVERSE64(&iBitPosition, pTDepth->bbField[BB_WHITEMAN]) ;
	iWhiteSlice = ( iBitScan ? m_iWhiteSlice[iBitPosition] : 0 ) ;

	ASSERT ( iWhiteSlice >= 0 && iWhiteSlice <= 40 ) ;

	iBitScan	= _BitScanForward64(&iBitPosition, pTDepth->bbField[BB_BLACKMAN]) ;
	iBlackSlice = ( iBitScan ? m_iBlackSlice[iBitPosition] : 0 ) ;

	ASSERT ( iBlackSlice >= 0 && iBlackSlice <= 40 ) ;

	iMatVal		= ( pTDepth->bTurn ? pTDepth->matval/10 + ( iWhiteSlice - iBlackSlice ) : -pTDepth->matval/10 - ( iWhiteSlice - iBlackSlice ) ) ;

	if ( iDB64Score == DB64_WIN ) {		// DB64_WIN
		pTDepth->uEvalFlags = EVAL_EXACT_WIN ;
		*iScore = DATABASE_WIN_VAL - 50 * ( pTDepth->iwn + pTDepth->ibn ) + iMatVal ;

	} else {							// DB64_LOSE
		pTDepth->uEvalFlags = EVAL_EXACT_LOOSE ;
		*iScore = DATABASE_LOOSE_VAL + 50 * ( pTDepth->iwn + pTDepth->ibn ) + iMatVal ;
	}

	return ( true ) ;

}	// End Database Lookup


int CDatabase64::DatabasePositionDB64(PTSTACK pTDepth)
{
	int iValue ;


	// Endgame databases ( if total pieces W + B <= DB64CACHE_MAXPIECE )
	if ( ( ( pTDepth->iwn + pTDepth->ibn ) > DB64CACHE_MAXPIECE ) )
		return ( DB64_UNKNOWN ) ;

	iValue = DB64Cache_Read(pTDepth, DB64_COMPRESS_DEFAULT) ;

	return ( iValue ) ;
}
